#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

# pylint: disable=too-few-public-methods
class common(object):
    """ json exporter plugin.
        By convention it has to be called exactly the same as its file name.
        (Apart from .py extention)
    """

    def __init__(self):
        """ constructor """
        pass

    @staticmethod
    def extract_message_data(msg):
        """ Extracts user name from 'sender', message caption and message content from msg.
            :param msg: Raw message object.

            :return
                (...) tuple of message attributes
        """
        sender = msg.sender

        # Get the name of the sender if any
        is_sent_by_bot = None
        if sender:
            name = getattr(sender, 'username', None)
            if not name:
                name = getattr(sender, 'title', None)
                if not name:
                    name = (sender.first_name or "") + " " + (sender.last_name or "")
                    name = name.strip()
                if not name:
                    name = '???'
            is_sent_by_bot = getattr(sender, 'bot', None)
        else:
            name = '???'

        caption = None
        if hasattr(msg, 'message'):
            content = msg.message
        elif hasattr(msg, 'action'):
            content = str(msg.action)
        else:
            # Unknown message, simply print its class name
            content = type(msg).__name__

        re_id_str = ''
        if hasattr(msg, 'reply_to_msg_id') and msg.reply_to_msg_id is not None:
            re_id_str = str(msg.reply_to_msg_id)

        is_contains_media = False
        media_content = None
        # Format the message content
        if getattr(msg, 'media', None):
            # The media may or may not have a caption
            is_contains_media = True
            caption = getattr(msg.media, 'caption', '')
            media_content = '<{}> {}'.format(
                type(msg.media).__name__, caption)

        return name, caption, content, re_id_str, is_sent_by_bot, is_contains_media, media_content



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

import json
from datetime import date, datetime
from .common import common

class jsonl(object):
    """ jsonl exporter plugin.

    As opposed to json exporter jsonl serializes messages as one JSON object per line, not as
    one giant array.

    See http://jsonlines.org.
    """
    # pylint: disable=no-self-use

    def __init__(self):
        """ constructor """
        pass

    # pylint: disable=unused-argument
    def format(self, msg, exporter_context):
        """ Formatter method. Takes raw msg and converts it to a *one-line* string.
            :param msg: Raw message object :class:`telethon.tl.types.Message` and derivatives.
                        https://core.telegram.org/type/Message

            :returns: *one-line* string containing one message data.
        """
        # pylint: disable=line-too-long
        name, _, content, re_id, is_sent_by_bot, is_contains_media, media_content = common.extract_message_data(msg)

        msgDictionary = {
            'message_id': msg.id,
            'from_id': msg.from_id,
            'reply_id': re_id,
            'author': name,
            'sent_by_bot': is_sent_by_bot,
            'date': msg.date,
            'content': content,
            'contains_media': is_contains_media,
            'media_content': media_content
        }
        msg_dump_str = json.dumps(
            msgDictionary, default=self._json_serial, ensure_ascii=False)
        return msg_dump_str

    def begin_final_file(self, resulting_file, exporter_context):
        """ Hook executes at the beginning of writing a resulting file.
            (After BOM is written in case of --addbom)
        """
        pass

    def _json_serial(self, obj):
        """JSON serializer for objects not serializable by default json code
           https://stackoverflow.com/questions/11875770/how-to-overcome-datetime-datetime-not-json-serializable
        """
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError("Type %s not serializable" % type(obj))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" This Module contains classes related to Metadata Files"""

import os.path
import errno
import codecs
import json
import logging
from telegram_messages_dump.exceptions import MetadataError


class DumpMetadata:
    """ Metadata file CRUD """

    CHAT_NAME = "chat_name"
    LAST_MESSAGE_ID = "latest_message_id"
    EXPORTER = "exporter_name"

    def __init__(self, out_file_path):
        self.meta_file_path = out_file_path + '.meta'
        self._meta_dict = {}
        self.logger = logging.getLogger(__name__)

    def merge_into_settings(self, settings):
        """ Get exporter name from metadata"""
        if not self._meta_dict:
            self._loadFromFile()

        settings.chat_name = self._meta_dict[DumpMetadata.CHAT_NAME]
        settings.last_message_id = self._meta_dict[DumpMetadata.LAST_MESSAGE_ID]
        settings.exporter = self._meta_dict[DumpMetadata.EXPORTER]

    def _loadFromFile(self):
        """ Loads metadata from file """
        try:
            self.logger.debug('Load metafile %s.', self.meta_file_path)
            with codecs.open(self.meta_file_path, 'r', 'utf-8') as meta_file:
                self._meta_dict = json.load(meta_file)
                # TODO Validate Meta Dict
        except OSError as ex:
            raise MetadataError('Unable to open the metadata file "{}". {}'
                                .format(self.meta_file_path, ex.strerror)) from ex
        except ValueError as ex:
            raise MetadataError(
                'Unable to load the metadata file "{}". AttributeError: {}'
                .format(self.meta_file_path, ex)) from ex

    def delete_meta_file(self):
        """ Delete metafile if running in CONTINUE mode """
        try:
            self.logger.debug('Delete old metadata file %s.', self.meta_file_path)
            os.remove(self.meta_file_path)
        except OSError as ex:
            if ex.errno != errno.ENOENT:
                raise MetadataError('Failed to delete old metadata file. {}'
                                    .format(ex.strerror))

    def save_meta_file(self, new_dict):
        """ Save metafile to file"""
        try:
            self.logger.debug('Save new metadata file %s.', self.meta_file_path)
            if not self._meta_dict:
                self._meta_dict = {}

            self._meta_dict["schema"] = "http://telegram-messages-dump/schema/v/1"

            if DumpMetadata.CHAT_NAME in new_dict:
                self._meta_dict[DumpMetadata.CHAT_NAME] = new_dict[DumpMetadata.CHAT_NAME]
            if DumpMetadata.LAST_MESSAGE_ID in new_dict:
                self._meta_dict[DumpMetadata.LAST_MESSAGE_ID] =\
                     new_dict[DumpMetadata.LAST_MESSAGE_ID]
            if DumpMetadata.EXPORTER in new_dict:
                self._meta_dict[DumpMetadata.EXPORTER] = new_dict[DumpMetadata.EXPORTER]

            self.logger.info('Writing a new metadata file.')
            with open(self.meta_file_path, 'w') as mf:
                json.dump(self._meta_dict, mf, indent=4, sort_keys=False)
        except OSError as ex:
            raise MetadataError(
                'Failed to write the metadata file. {}'.format(ex.strerror))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

import re
from .common import common

class csv(object):
    """ csv (comma separated values) exporter plugin.
        By convention it has to be called exactly the same as its file name.
        (Apart from .py extention)
    """

    def __init__(self):
        """ constructor """
        self.ESCAPE = re.compile(r'[\x00-\x1f\\"\b\f\n\r\t]')
        self.ESCAPE_DICT = {
            '\\': '\\\\',
            '"': '""',
            '\b': '\\b',
            '\f': '\\f',
            '\n': '\\n',
            '\r': '\\r',
            '\t': '\\t',
        }
        for i in range(0x20):
            self.ESCAPE_DICT.setdefault(chr(i), '\\u{0:04x}'.format(i))

    def format(self, msg, exporter_context):
        """ Formatter method. Takes raw msg and converts it to a *one-line* string.
            :param msg: Raw message object :class:`telethon.tl.types.Message` and derivatives.
                        https://core.telegram.org/type/Message

            :returns: *one-line* string containing one message data.
        """
        # pylint: disable=unused-argument
        name, _, content, re_id, _, _, _ = common.extract_message_data(msg)
        # Format a message log record
        # msg_dump_str = '[{}-{:02d}-{:02d} {:02d}:{:02d}] ID={} {}{}: {}'.format(
        #     msg.date.year, msg.date.month, msg.date.day,
        #     msg.date.hour, msg.date.minute, msg.id, "RE_ID=%s " % re_id if re_id else "",
        #     name, self._py_encode_basestring(content))

        name, isNameModified = self._py_encode_basestring(name)

        # Check if name contains ',' or there were other special chars.
        # If it does surround the name with quotes.
        if name and (name.find(",") != -1 or isNameModified):
            name = '"' + name + '"'

        msg_dump_str = ",".join([str(msg.id),
                                 msg.date.isoformat(),
                                 name,
                                 str(re_id),
                                 '"' + str(self._py_encode_basestring(content)[0]) + '"'])
        return msg_dump_str

    def begin_final_file(self, resulting_file, exporter_context):
        """ Hook executes at the beginning of writing a resulting file.
            (After BOM is written in case of --addbom)
        """
        if not exporter_context.is_continue_mode:
            header_str = ",".join(["Message Id", "Time", "Sender Name", "Reply Id", "Message"])
            print(header_str, file=resulting_file)

    # This code is inspired by Python's json encoder's code
    def _py_encode_basestring(self, s):
        """Return a JSON representation of a Python string"""
        if not s:
            return s, False
        isAnyCharReplaced = False
        def replace(match):
            nonlocal isAnyCharReplaced
            isAnyCharReplaced = True
            return self.ESCAPE_DICT[match.group(0)]
        return str(self.ESCAPE.sub(replace, s)), isAnyCharReplaced



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Usage:

Normal mode:
  telegram-messages-dump -c <chat_name> -p <phone_num> [-l <count>] [-o <file>] [-cl] [...]
  telegram-messages-dump --chat=<chat_name> --phone=<phone_num> [--limit=<count>] [--out <file>]

Continuous mode:
  telegram-messages-dump --continue -p <phone_num> -o <file> [-cl] [...]
  telegram-messages-dump --continue=<MSG_ID> -p <phone_num> -o <file> -e <exporter> -c <chat_name>

Where:
    -c,  --chat      Unique name of a channel/chat. E.g. @python.
    -p,  --phone     Phone number. E.g. +380503211234.
    -o,  --out       Output file name or full path. (Default: telegram_<chatName>.log)
    -e,  --exp       Exporter name. text | jsonl | csv (Default: 'text')
      ,  --continue  Continue previous dump. Supports optional integer param <message_id>.
    -l,  --limit     Number of the latest messages to dump, 0 means no limit. (Default: 100)
    -cl, --clean     Clean session sensitive data (e.g. auth token) on exit. (Default: False)
    -v,  --verbose   Verbose mode. (Default: False)
      ,  --addbom    Add BOM to the beginning of the output file. (Default: False)
    -h,  --help      Show this help message and exit.
"""

import os
import sys
import importlib
import logging
from telegram_messages_dump.telegram_dumper import TelegramDumper
from telegram_messages_dump.chat_dump_settings import ChatDumpSettings
from telegram_messages_dump.chat_dump_metadata import DumpMetadata
from telegram_messages_dump.chat_dump_metadata import MetadataError
from telegram_messages_dump.utils import sprint

def main():
    """ Entry point. """
    settings = ChatDumpSettings(__doc__)

    # define the console output verbosity
    default_format = '%(levelname)s:%(message)s'
    if settings.is_verbose:
        logging.basicConfig(format=default_format, level=logging.DEBUG)
    else:
        logging.basicConfig(format=default_format, level=logging.INFO)

    metadata = DumpMetadata(settings.out_file)

    # when user specified --continue
    try:
        if settings.is_incremental_mode and settings.last_message_id == -1:
            metadata.merge_into_settings(settings)
    except MetadataError as ex:
        sprint("ERROR: %s" % ex)
        sys.exit(1)

    exporter = _load_exporter(settings.exporter)

    sys.exit(TelegramDumper(os.path.basename(__file__), settings, metadata, exporter).run())

def _load_exporter(exporter_name):
    """ Loads exporter from file <exporter_name>.py in ./exporters subfolder.
        :param exporter_name:      name of exporter. E.g. 'text' or 'json'

        :return: Exporter instance
    """
    # By convention exporters are located in .\exporters subfolder
    # COMMENT: Don't check file existance. It won't play well with pyinstaller bins
    exporter_file_name = exporter_name + ".py"
    exporter_rel_name = "telegram_messages_dump.exporters." + exporter_name
    # Load exporter from file
    sprint("Try to load exporter '%s'...  " % (exporter_file_name), end='')
    try:
        exporter_module = importlib.import_module(exporter_rel_name)
        sprint("OK!")
    except ModuleNotFoundError:
        sprint("\nERROR: Failed to load exporter './exporters/%s'." % exporter_file_name)
        exit(1)

    try:
        exporterClass = getattr(exporter_module, exporter_name)
    except AttributeError:
        sprint("ERROR: Failed to load class '%s' out of './exporters/%s'." \
               % (exporter_name, exporter_file_name))
        exit(1)

    return exporterClass()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Various utility functions/classes """

def sprint(string, *args, **kwargs):
    """Safe Print (handle UnicodeEncodeErrors on some terminals)"""
    try:
        print(string, *args, **kwargs)
    except UnicodeEncodeError:
        string = string.encode('utf-8', errors='ignore') \
            .decode('ascii', errors='ignore')
        print(string, *args, **kwargs)


JOIN_CHAT_PREFIX_URL = 'https://t.me/joinchat/'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" This Module contains classes related to CLI interactions"""

import argparse
from telegram_messages_dump.utils import JOIN_CHAT_PREFIX_URL


class ChatDumpSettings:
    """ Parses CLI arguments. """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods
    def __init__(self, usage):

        # From telegram-cli
        self.api_id = 2899
        self.api_hash = '36722c72256a24c1225de00eb6a1ca74'

        # Parse parameters
        parser = CustomArgumentParser(
            formatter_class=CustomFormatter, usage=usage)

        parser.add_argument('-c', '--chat', default='', required=False, type=str)
        parser.add_argument('-p', '--phone', required=True, type=str)
        parser.add_argument('-o', '--out', default='', type=str)
        parser.add_argument('-e', '--exp', default='', type=str)
        parser.add_argument('--continue', dest='increment',
                            default='*', type=str, nargs='?')
        parser.add_argument('-l', '--limit', default=-1, type=int)
        parser.add_argument('-cl', '--clean', action='store_true')
        parser.add_argument('-v', '--verbose', action='store_true')
        parser.add_argument('--addbom', action='store_true')
        parser.add_argument('-q', '--quiet', action='store_true')

        args = parser.parse_args()

        # Trim extra spaces in string param values
        if args.chat:
            args.chat = args.chat.strip()
        if args.exp:
            args.exp = args.exp.strip()
        if args.out:
            args.out = args.out.strip()
        if args.phone:
            args.phone = args.phone.strip()

        # Detect Normal/Incremental mode
        self._process_incremental_mode_option(args, parser)

        # Check if user specified the right options depending on the mode
        self._check_options_consistency(args, parser)

        # Validate chat name
        if self.last_message_id != -1 and not args.chat != "":
            parser.error("Chat name can't be empty")

        # Validate phone number
        try:
            if int(args.phone) <= 0:
                raise ValueError
        except ValueError:
            parser.error('Phone number is invalid.')

        # Validate limit / set default
        if not self.is_incremental_mode and args.limit < 0:
            args.limit = 100

        # Validate exporter name / set default
        exp_file = 'text' if not args.exp else args.exp
        if not exp_file:
            parser.error('Exporter name is invalid.')

        # Default output file if not specified by user
        OUTPUT_FILE_TEMPLATE = 'telegram_{}.log'
        if args.out != '':
            out_file = args.out
        elif args.chat.startswith(JOIN_CHAT_PREFIX_URL):
            out_file = OUTPUT_FILE_TEMPLATE.format(args.chat.rsplit('/', 1)[-1])
        else:
            out_file = OUTPUT_FILE_TEMPLATE.format(args.chat)

        self.chat_name = args.chat
        self.phone_num = args.phone
        self.out_file = out_file
        self.exporter = exp_file
        self.limit = args.limit
        self.is_clean = args.clean
        self.is_verbose = args.verbose
        self.is_addbom = args.addbom
        self.is_quiet_mode = args.quiet

    def _process_incremental_mode_option(self, args, parser):
        """ Arguments parsing related to --continue setting """
        self.last_message_id = -1
        self.is_incremental_mode = False

        if not args.increment == '*':
            # if user specified --continue with/without a parameter
            self.is_incremental_mode = True
            if args.increment:
                # else-if user specified --continue with one param that is expected to be uint
                # Try parse it as int assuming it is the last message_id
                try:
                    self.last_message_id = int(args.increment)
                except ValueError:
                    parser.error(
                        'Unable to parse MSG_ID in --continue=<MSG_ID>')
        return

    def _check_options_consistency(self, args, parser):
        if self.is_incremental_mode:
            if args.out == "":
                parser.error('To increment an existing dump file. '
                             'You have to specify it using --out or -o setting.')
            if self.last_message_id != -1:
                # In case of --continue=<MSG_ID>
                if args.chat == "":
                    parser.error('chat name must be specified explicitely '
                                 'when using --continue=<MSG_ID>')
                if args.exp == "":
                    parser.error('exporter must be specified explicitely '
                                 'when using --continue=<MSG_ID>')
                if args.limit != -1:
                    parser.error('limit setting is not allowed '
                                 'when using --continue=<MSG_ID>')
            else:
                # In case of --continue
                if args.chat != "":
                    parser.error('chat name must NOT be specified explicitely '
                                 'when using --continue')
                if args.exp != "":
                    parser.error('exporter must NOT be specified explicitely '
                                 'when using --continue')
                if args.limit != -1:
                    parser.error('limit setting is not allowed '
                                 'when using --continue')
        else:
            # In case of Normal mode
            if args.chat == "":
                parser.error('the following arguments are required: -c/--chat ')
        return


class CustomFormatter(argparse.HelpFormatter):
    """ Custom formatter for setting argparse formatter_class.
        It only outputs raw 'usage' text and omits other sections
        (e.g. positional, optional params and epilog).
    """

    def __init__(self, prog=''):
        argparse.HelpFormatter.__init__(
            self, prog, max_help_position=100, width=150)

    def add_usage(self, usage, actions, groups, prefix=None):
        if usage is not argparse.SUPPRESS:
            args = usage, actions, groups, ''
            self._add_item(self._format_usage, args)

    def _format_usage(self, usage, actions, groups, prefix):
        # if usage is specified, use that
        if usage is not None:
            usage = usage % dict(prog=self._prog)

        return "\n\r%s\n\r" % usage


class CustomArgumentParser(argparse.ArgumentParser):
    """ Custom ArgumentParser.
        Outputs raw 'usage' text and omits other sections.
    """

    def format_help(self):
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        return formatter.format_help()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" This is a main Module that contains code
    that fetches messages from Telegram and do processing.
"""

import os
import os.path
import sys
import codecs
import tempfile
import logging
from collections import deque
from getpass import getpass
from time import sleep
from telethon import TelegramClient, sync # pylint: disable=unused-import
from telethon.errors import (FloodWaitError,
                             SessionPasswordNeededError,
                             UsernameNotOccupiedError,
                             UsernameInvalidError)
from telethon.tl.functions.contacts import ResolveUsernameRequest
from telegram_messages_dump.utils import sprint
from telegram_messages_dump.utils import JOIN_CHAT_PREFIX_URL
from telegram_messages_dump.exceptions import DumpingError
from telegram_messages_dump.exceptions import MetadataError
from telegram_messages_dump.exporter_context import ExporterContext


class TelegramDumper(TelegramClient):
    """ Authenticates and opens new session. Retrieves message history for a chat. """

    def __init__(self, session_user_id, settings, metadata, exporter):

        self.logger = logging.getLogger(__name__)
        self.logger.info('Initializing session...')
        super().__init__(session_user_id,
                         settings.api_id,
                         settings.api_hash,
                         timeout=40, #seconds
                         proxy=None)

        # Settings as specified by user or defaults or from metadata
        self.settings = settings

        # Metadata that was possibly loaded from .meta file and will be saved there
        self.metadata = metadata

        # Exporter object that converts msg -> string
        self.exporter = exporter

        # The context that will be passed to the exporter
        self.exporter_context = ExporterContext()
        self.exporter_context.is_continue_mode = self.settings.is_incremental_mode

        # How many massages user wants to be dumped
        # explicit --limit, or default of 100 or unlimited (int.Max)
        self.msg_count_to_process = 0

        # Messages page offset for fetching
        self.id_offset = 0

        # A list of paths to the temp files
        self.temp_files_list = deque()

        # Actual lattets message id that was prossessed since the dumper started running
        self.cur_latest_message_id = self.settings.last_message_id

        # The number of messages written into a resulting file de-facto
        self.output_total_count = 0

    def run(self):
        """ Dumps all desired chat messages into a file """

        ret_code = 0
        try:
            self._init_connect()
            try:
                chatObj = self._getChannel()
            except ValueError as ex:
                ret_code = 1
                self.logger.error('%s', ex,
                                  exc_info=self.logger.level > logging.INFO)
                return
            # Fetch history in chunks and save it into a resulting file
            self._do_dump(chatObj)
        except (DumpingError, MetadataError) as ex:
            self.logger.error('%s', ex, exc_info=self.logger.level > logging.INFO)
            ret_code = 1
        except KeyboardInterrupt:
            sprint("Received a user's request to interrupt, stopping…")
            ret_code = 1
        except Exception as ex:  # pylint: disable=broad-except
            self.logger.error('Uncaught exception occured. %s', ex,
                              exc_info=self.logger.level > logging.INFO)
            ret_code = 1
        finally:
            self.logger.debug('Make sure there are no temp files left undeleted.')
            # Clear temp files if any
            while self.temp_files_list:
                try:
                    os.remove(self.temp_files_list.pop().name)
                except Exception:  # pylint: disable=broad-except
                    pass

        if self.settings.is_clean:
            try:
                # TODO
                # self.log_out()
                self.logger.info('Session data cleared.')
            except Exception:  # pylint: disable=broad-except
                sprint('Failed to logout and clean session data.')

        sprint('{} messages were successfully written in the resulting file. Done!'
               .format(self.output_total_count))
        return ret_code

    def _init_connect(self):
        """ Connect to the Telegram server and Authenticate. """
        sprint('Connecting to Telegram servers...')
        if not self.connect():
            sprint('Initial connection failed.')

        # Then, ensure we're authorized and have access
        if not self.is_user_authorized():
            sprint('First run. Sending code request...')
            self.send_code_request(self.settings.phone_num)
            self_user = None
            while self_user is None:
                code = input('Enter the code you just received: ')
                try:
                    self_user = self.sign_in(self.settings.phone_num, code)
                # Two-step verification may be enabled
                except SessionPasswordNeededError:
                    pw = getpass("Two step verification is enabled. "
                                 "Please enter your password: ")
                    self_user = self.sign_in(password=pw)

    def _getChannel(self):
        """ Returns telethon.tl.types.Channel object resolved from chat_name
            at Telegram server
        """
        name = self.settings.chat_name

        # For private channуls try to resolve channel peer object from its invitation link
        # Note: it will only work if the login user has already joined the private channel.
        # Otherwise, get_entity will throw ValueError
        if name.startswith(JOIN_CHAT_PREFIX_URL):
            self.logger.debug('Trying to resolve as invite url.')
            try:
                peer = self.get_entity(name)
                if peer:
                    sprint('Invitation link "{}" resolved into channel id={}'.format(
                        name, peer.id))
                    return peer
            except ValueError as ex:
                self.logger.debug('Failed to resolve "%s" as an invitation link. %s',
                                  self.settings.chat_name,
                                  ex,
                                  exc_info=self.logger.level > logging.INFO)

        if name.startswith('@'):
            name = name[1:]
            self.logger.debug('Trying ResolveUsernameRequest().')
            try:
                peer = self(ResolveUsernameRequest(name))
                if peer.chats is not None and peer.chats:
                    sprint('Chat name "{}" resolved into channel id={}'.format(
                        name, peer.chats[0].id))
                    return peer.chats[0]
                if peer.users is not None and peer.users:
                    sprint('User name "{}" resolved into channel id={}'.format(
                        name, peer.users[0].id))
                    return peer.users[0]
            except (UsernameNotOccupiedError, UsernameInvalidError) as ex:
                self.logger.debug('Failed to resolve "%s" as @-chat-name. %s',
                                  self.settings.chat_name,
                                  ex,
                                  exc_info=self.logger.level > logging.INFO)

        # Search in dialogs first, this way we will find private groups and
        # channels.
        self.logger.debug('Fetch loggedin user`s dialogs')
        dialogs_count = self.get_dialogs(0).total
        self.logger.info('%s user`s dialogs found', dialogs_count)
        dialogs = self.get_dialogs(limit=None)
        self.logger.debug('%s dialogs fetched.', len(dialogs))
        for dialog in dialogs:
            if dialog.name == name:
                sprint('Dialog title "{}" resolved into channel id={}'.format(
                    name, dialog.entity.id))
                return dialog.entity
            if hasattr(dialog.entity, 'username') and dialog.entity.username == name:
                sprint('Dialog username "{}" resolved into channel id={}'.format(
                    name, dialog.entity.id))
                return dialog.entity
            if name.startswith('@') and dialog.entity.username == name[1:]:
                sprint('Dialog username "{}" resolved into channel id={}'.format(
                    name, dialog.entity.id))
                return dialog.entity
        self.logger.debug('Specified chat name was not found among dialogs.')

        raise ValueError('Failed to resolve dialogue/chat name "{}".'.format(name))

    def _fetch_messages_from_server(self, peer, buffer):
        """ Retrieves a number (100) of messages from Telegram's DC and adds them to 'buffer'.
            :param peer:        Chat/Channel object
            :param buffer:      buffer where to place retrieved messages

            :return The latest/biggest Message ID that successfully went into buffer,
                    or -1 if there are no more messages.
        """
        messages = []

        # First retrieve the messages and some information
        # make 5 attempts
        for _ in range(0, 5):
            try:
                # NOTE: Telethon will make 5 attempts to reconnect
                # before failing
                messages = self.get_messages(
                    peer, limit=100, offset_id=self.id_offset)

                if messages.total > 0 and messages:
                    sprint('Processing messages with ids {}-{} ...'
                           .format(messages[0].id, messages[-1].id))
            except FloodWaitError as ex:
                sprint('FloodWaitError detected. Sleep for {} sec before reconnecting! \n'
                       .format(ex.seconds))
                sleep(ex.seconds)
                self._init_connect()
                continue
            break

        latest_message_id = -1 \
            if not messages or self.settings.last_message_id >= messages[0].id \
            else messages[0].id

        # Iterate over all (in reverse order so the latest appear
        # the last in the console) and print them with format provided by exporter.
        for msg in messages:
            self.exporter_context.is_first_record = \
                self.msg_count_to_process == 1

            if self.settings.last_message_id >= msg.id:
                self.msg_count_to_process = 0
                break

            msg_dump_str = self.exporter.format(msg, self.exporter_context)

            buffer.append(msg_dump_str)

            self.msg_count_to_process -= 1
            self.id_offset = msg.id
            self.exporter_context.is_last_record = False
            if self.msg_count_to_process == 0:
                break

        return latest_message_id

    def _do_dump(self, peer):
        """ Retrieves messages in small chunks (Default: 100) and saves them in in-memory 'buffer'.
            When buffer reaches '1000' messages they are saved into intermediate temp file.
            In the end messages from all the temp files are being moved into resulting file in
            ascending order along with the remaining ones in 'buffer'.
            After all, temp files are deleted.

             :param peer: Chat/Channel object that contains the message history of interest

             :return  Number of files that were saved into resulting file
        """
        self.msg_count_to_process = self.settings.limit \
            if self.settings.limit != -1\
            and not self.settings.limit == 0\
            and not self.settings.is_incremental_mode\
            else sys.maxsize

        self._check_preconditions()

        # Current buffer of messages, that will be batched into a temp file
        # or otherwise written directly into the resulting file if there are too few of them
        # to form a batch of size 1000.
        buffer = deque()

        # Delete old metafile in Continue mode
        if not self.settings.is_incremental_mode:
            self.metadata.delete_meta_file()

        temp_files_list_meta = deque()  # a list of meta info about batches

        # process messages until either all message count requested by user are retrieved
        # or offset_id reaches msg_id=1 - the head of a channel message history
        try:
            while self.msg_count_to_process > 0:
                # slip for a few seconds to avoid flood ban
                sleep(2)

                latest_message_id_fetched = self._fetch_messages_from_server(
                    peer, buffer)

                # This is for the case when buffer with fewer than 1000 records
                # Relies on the fact that `_fetch_messages_from_server` returns messages
                # in reverse order
                if self.cur_latest_message_id < latest_message_id_fetched:
                    self.cur_latest_message_id = latest_message_id_fetched
                # when buffer is full, flush it into a temp file
                # Assume that once a message got into temp file it will be counted as successful
                # 'output_total_count'. This has to be improved.
                if len(buffer) >= 1000:
                    self._flush_buffer_in_temp_file(buffer)
                    temp_files_list_meta.append(latest_message_id_fetched)

                # break if the very beginning of channel history is reached
                if latest_message_id_fetched == -1 or self.id_offset <= 1:
                    break
        except RuntimeError as ex:
            sprint('Fetching messages from server failed. ' + str(ex))
            sprint('Warn: The resulting file will contain partial/incomplete data.')

        # Write all chunks into resulting file
        sprint('Merging results into an output file.')
        try:
            self._write_final_file(buffer, temp_files_list_meta)
        except OSError as ex:
            raise DumpingError("Dumping to a final file failed.") from ex

        # Metadata that will be written into a metafile
        meta_dict = {
            "latest_message_id": self.cur_latest_message_id,
            "exporter_name": self.settings.exporter,
            "chat_name": self.settings.chat_name
        }
        self.metadata.save_meta_file(meta_dict)

    def _flush_buffer_in_temp_file(self, buffer):
        """ Flush buffer into a new temp file """
        with tempfile \
                .NamedTemporaryFile(mode='w+', encoding='utf-8', delete=False) as tf:
            self.output_total_count += self._flush_buffer_into_filestream(buffer, tf)
            self.temp_files_list.append(tf)

    def _flush_buffer_into_filestream(self, buffer, file_stream):
        """ Flush buffer into a file stream """
        count = 0
        while buffer:
            count += 1
            cur_message = buffer.pop()
            print(cur_message, file=file_stream)
        return count

    def _write_final_file(self, buffer, temp_files_list_meta):
        result_file_mode = 'a' if self.settings.last_message_id > -1 else 'w'
        with codecs.open(self.settings.out_file, result_file_mode, 'utf-8') as resulting_file:
            if self.settings.is_addbom:
                resulting_file.write(codecs.BOM_UTF8.decode())

            self.exporter.begin_final_file(
                resulting_file, self.exporter_context)

            # flush what's left in the mem buffer into resulting file
            self.output_total_count += self._flush_buffer_into_filestream(
                buffer, resulting_file)

            self._merge_temp_files_into_final(
                resulting_file, temp_files_list_meta)

    def _merge_temp_files_into_final(self, resulting_file, temp_files_list_meta):
        """ merge all temp files into final one and delete them """
        while self.temp_files_list:
            tf = self.temp_files_list.pop()
            with codecs.open(tf.name, 'r', 'utf-8') as ctf:
                for line in ctf.readlines():
                    print(line, file=resulting_file, end='')
            # delete temp file
            self.logger.debug("Delete temp file %s", tf.name)
            tf.close()
            os.remove(tf.name)
            # update the latest_message_id metadata
            batch_latest_message_id = temp_files_list_meta.pop()
            if batch_latest_message_id > self.cur_latest_message_id:
                self.cur_latest_message_id = batch_latest_message_id

    def _check_preconditions(self):
        """ Check preconditions before processing data """
        out_file_path = self.settings.out_file
        if self.settings.is_incremental_mode:
            # In incrimental mode
            sprint('Switching to incremental mode.')
            self.logger.debug('Checking if output file exists.')
            if not os.path.exists(out_file_path):
                raise DumpingError(
                    'Error: Output file does not exist. Path="' + out_file_path + '"')
            sprint('Dumping messages newer than {} using "{}" dumper.'
                   .format(self.settings.last_message_id, self.settings.exporter))
        else:
            # In NONE-incrimental mode
            if os.path.exists(out_file_path):
                sprint('Warning: The output file already exists.')
                if not self._is_user_confirmed('Are you sure you want to overwrite it? [y/n]'):
                    raise DumpingError("Terminating on user's request...")
            # Check if output file can be created/overwritten
            try:
                with open(out_file_path, mode='w+'):
                    pass
            except OSError as ex:
                raise DumpingError('Output file path "{}" is invalid. {}'.format(
                    out_file_path, ex.strerror))
            sprint('Dumping {} messages into "{}" file ...'
                   .format('all' if self.msg_count_to_process == sys.maxsize
                           else self.msg_count_to_process, out_file_path))

    def _is_user_confirmed(self, msg):
        """ Get confirmation from user """
        if self.settings.is_quiet_mode:
            return True
        continueResponse = input(msg).lower().strip()
        return continueResponse == 'y'\
            or continueResponse == 'yes'



""" Custom Exceptions used throughout telegram_message_dump tool. """

class DumpingError(Exception):
    """ Dumping exception"""
    pass


class MetadataError(Exception):
    """ Metadata processing exception"""
    pass




"""Setup script for telegram-messages-dump"""

import os.path
from setuptools import setup
from telegram_messages_dump import __version__

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

setup(
    name='telegram-messages-dump',
    version=__version__,
    description='Telegram Messages Dump console utility.',
    long_description_content_type="text/markdown",
    long_description=README,
    author='Kostya Kravets',
    author_email='kostikkv@gmail.com',
    url='https://github.com/Kosat/telegram-messages-dump',
    download_url='https://github.com/Kosat/telegram-messages-dump/releases',
    install_requires=['telethon==1.6.2'],
    license="MIT",
    packages=['telegram_messages_dump', "telegram_messages_dump.exporters"],
    include_package_data=True,
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3'
    ],
    entry_points={
        'console_scripts':
            ['telegram-messages-dump = telegram_messages_dump.run:main']
    }
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

class ExporterContext:
    """ Exporter context """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods
    def __init__(self):
        # Is processing the first record
        self.is_first_record = False
        # Is processing the last record
        self.is_last_record = True
        # Is working in continue/incremental mode
        self.is_continue_mode = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring

import re
from .common import common

class text(object):
    """ text exporter plugin.
        By convention it has to be called exactly the same as its file name.
        (Apart from .py extention)
    """

    def __init__(self):
        """ constructor """
        self.ESCAPE = re.compile(r'[\x00-\x1f\b\f\n\r\t]')
        self.ESCAPE_DICT = {
            '\\': '\\\\',
            # '"': '\\"',
            '\b': '\\b',
            '\f': '\\f',
            '\n': '\\n',
            '\r': '\\r',
            '\t': '\\t',
        }
        for i in range(0x20):
            self.ESCAPE_DICT.setdefault(chr(i), '\\u{0:04x}'.format(i))

    def format(self, msg, exporter_context):
        """ Formatter method. Takes raw msg and converts it to a *one-line* string.
            :param msg: Raw message object :class:`telethon.tl.types.Message` and derivatives.
                        https://core.telegram.org/type/Message

            :returns: *one-line* string containing one message data.
        """
        # pylint: disable=unused-argument
        name, _, content, re_id, _, _, _ = common.extract_message_data(msg)
        # Format a message log record
        msg_dump_str = '[{}-{:02d}-{:02d} {:02d}:{:02d}] ID={} {}{} {}: {}'.format(
            msg.date.year, msg.date.month, msg.date.day,
            msg.date.hour, msg.date.minute, msg.id, "RE_ID=%s " % re_id if re_id else "",
            name, getattr(msg, "sender_id", None), self._py_encode_basestring(content))

        return msg_dump_str

    def begin_final_file(self, resulting_file, exporter_context):
        """ Hook executes at the beginning of writing a resulting file.
            (After BOM is written in case of --addbom)
        """
        pass

    # This code is inspired by Python's json encoder's code
    def _py_encode_basestring(self, s):
        """Return a JSON representation of a Python string"""
        if not s:
            return s
        def replace(match):
            return self.ESCAPE_DICT[match.group(0)]
        return self.ESCAPE.sub(replace, s)
